package main

import (
	"errors"
	"flag"
	"fmt"
	"log"
	"strings"

	bolt "go.etcd.io/bbolt"
)

// getBucket splits the given path using / as the separator and returns a nested bucket with the given path, if it exists.
func getBucket(tx *bolt.Tx, path string) *bolt.Bucket {
	parts := strings.Split(strings.Trim(path, "/"), "/")
	if len(parts) == 0 {
		return nil
	}

	bucket := tx.Bucket([]byte(parts[0]))
	parts = parts[1:]

	for len(parts) > 0 {
		bucket = bucket.Bucket([]byte(parts[0]))
		if bucket == nil {
			return nil
		}

		parts = parts[1:]
	}

	return bucket
}

func commandList(db *bolt.DB, args []string) error {
	if len(args) > 1 {
		return errors.New("too many arguments")
	}

	return db.View(func(tx *bolt.Tx) error {
		if len(args) == 0 {
			return tx.ForEach(func(k []byte, bucket *bolt.Bucket) error {
				fmt.Printf("b\t%q\n", k)

				return nil
			})
		}

		bucket := getBucket(tx, args[0])
		if bucket == nil {
			return errors.New("no such bucket")
		}

		return bucket.ForEach(func(k, v []byte) error {
			prefix := "-"
			if v == nil {
				prefix = "b"
			}

			fmt.Printf("%s\t%q\n", prefix, k)

			return nil
		})
	})
}

type bucketDeleter interface {
	DeleteBucket([]byte) error
	Bucket([]byte) *bolt.Bucket
}

func commandDeleteBucket(db *bolt.DB, args []string) error {
	if len(args) != 1 {
		return errors.New("expected exactly one argument")
	}

	path := strings.Split(strings.Trim(args[0], "/"), "/")

	fmt.Println("removing", path)

	return db.Update(func(tx *bolt.Tx) error {
		var del bucketDeleter

		del = tx

		for len(path) > 1 {
			del = del.Bucket([]byte(path[0]))
			if del == nil {
				return errors.New("not found")
			}

			path = path[1:]
		}

		return del.DeleteBucket([]byte(path[0]))
	})
}

func main() {
	dbPath := flag.String("dbPath", "../../vino.db", "Path to database file")

	flag.Parse()

	db, err := bolt.Open(*dbPath, 0644, nil)
	if err != nil {
		log.Fatalln("can't open DB:", err)
	}
	defer db.Close()

	command := "ls"
	if len(flag.Args()) > 0 {
		command = flag.Args()[0]
	}

	args := flag.Args()[1:]

	switch command {
	case "ls":
		err = commandList(db, args)
	case "del":
		err = commandDeleteBucket(db, args)
	default:
		err = fmt.Errorf("unexpected command %q", command)
	}

	if err != nil {
		log.Fatalln("can't view db:", err)
	}
}
