package main

import (
	"crypto/rand"
	"fmt"

	bolt "go.etcd.io/bbolt"
)

const expectDBVersion = "1"

// initDB initializes db with persistent state like the password salt and checks whether an already initialized database has the expected version number.
func initDB(db *bolt.DB) error {
	// Initialize persistent state like password salt
	err := db.Update(func(tx *bolt.Tx) error {
		bucket, err := tx.CreateBucketIfNotExists([]byte("meta"))
		if err != nil {
			return err
		}

		// Check database version
		storedVersion := bucket.Get([]byte("version"))
		if storedVersion != nil && string(storedVersion) != expectDBVersion {
			return fmt.Errorf("Unexpected database version: have %q, want %q", storedVersion, expectDBVersion)
		}

		// Make sure the db version is stored
		err = bucket.Put([]byte("version"), []byte(expectDBVersion))
		if err != nil {
			return err
		}

		// Create new PW salt if it doesn't already exist
		pwSalt := bucket.Get([]byte("pwsalt"))
		if pwSalt == nil {
			pwSalt = make([]byte, 16)
			_, err := rand.Read(pwSalt)
			if err != nil {
				return err
			}

			err = bucket.Put([]byte("pwsalt"), pwSalt)
			if err != nil {
				return err
			}

		}

		return nil
	})
	if err != nil {
		return err
	}

	return nil
}
