package main

import (
	"bytes"
	"context"
	"errors"
	"fmt"
	"image"
	"image/png"
	"io"
	"log"
	"strconv"

	// Imported for side effects to register format handlers
	_ "image/jpeg"
	_ "image/png"

	"github.com/google/uuid"
	bolt "go.etcd.io/bbolt"
	"golang.org/x/image/draw"
)

var errNotFound = errors.New("not found")

type Comment struct {
	ID      uuid.UUID
	Content string
}

type Vino struct {
	ID         uuid.UUID
	Name       string
	Rating     int
	Picture    image.Image
	HasPicture bool // Set to true if there's picture data for this vino
	Comments   []Comment
}

func DeleteVino(ctx context.Context, db *bolt.DB, id uuid.UUID) error {
	return db.Update(func(tx *bolt.Tx) error {
		allWines := tx.Bucket([]byte("wines"))
		if allWines == nil {
			return nil
		}

		return allWines.DeleteBucket([]byte(id.String()))
	})
}

// loadVino uses the given read-only bolt transaction to load the data for the wine with the given id. When
// there are no wines at all, or there is no wine with the given ID, loadVino returns errNotFound.
func loadVino(tx *bolt.Tx, id uuid.UUID) (Vino, error) {
	allWines := tx.Bucket([]byte("wines"))
	if allWines == nil {
		return Vino{}, errNotFound
	}

	data := allWines.Bucket([]byte(id.String()))
	if data == nil {
		return Vino{}, errNotFound
	}

	v := Vino{
		ID:         id,
		Name:       string(data.Get([]byte("name"))),
		HasPicture: data.Get([]byte("picture")) != nil, // TODO
	}

	var err error

	v.Rating, err = strconv.Atoi(string(data.Get([]byte("rating"))))
	if err != nil {
		return v, err
	}

	comments := data.Bucket([]byte("comments"))
	if comments != nil {
		err := comments.ForEach(func(k, txt []byte) error {
			id, err := uuid.ParseBytes(k)
			if err != nil {
				return err
			}

			v.Comments = append(v.Comments, Comment{
				ID:      id,
				Content: string(txt),
			})

			return nil
		})

		if err != nil {
			return v, err
		}
	}

	return v, nil
}

func LoadVino(ctx context.Context, db *bolt.DB, id uuid.UUID) (Vino, error) {
	var v Vino

	err := db.View(func(tx *bolt.Tx) error {
		var err error

		v, err = loadVino(tx, id)
		if err != nil {
			return err
		}

		return nil
	})

	if err != nil {
		return v, err
	}

	return v, nil
}

func ListWines(ctx context.Context, db *bolt.DB) ([]Vino, error) {
	var wines []Vino

	err := db.View(func(tx *bolt.Tx) error {
		bucket := tx.Bucket([]byte("wines"))
		if bucket == nil {
			return nil
		}

		return bucket.ForEach(func(k, v []byte) error {
			if v != nil {
				// We only expect nil values here because the elements in the bucket are all nested buckets
				panic(fmt.Sprintf("database corrupt, did not expect a value for key %q: %q", k, v))
			}

			id, err := uuid.ParseBytes(k)
			if err != nil {
				return fmt.Errorf("can't parse %q as UUID: %w", k, err)
			}

			vino, err := loadVino(tx, id)
			if err != nil {
				return err
			}

			wines = append(wines, vino)

			return nil
		})
	})

	if err != nil {
		return nil, err
	}

	return wines, nil
}

func LoadPictureData(ctx context.Context, db *bolt.DB, id uuid.UUID) ([]byte, error) {
	var data []byte

	err := db.View(func(tx *bolt.Tx) error {
		allWines := tx.Bucket([]byte("wines"))
		if allWines == nil {
			return errNotFound
		}

		wine := allWines.Bucket([]byte(id.String()))

		data = wine.Get([]byte("picture"))

		return nil
	})

	if err != nil {
		return nil, err
	}

	if len(data) == 0 {
		return nil, errNotFound
	}

	return data, nil
}

// AddPicture loads picture data (PNG or JPEG) from fh and sets v's picture to it.
// If something goes wrong during loading, or the image is neither PNG nor JPEG, an error
// is returned. If contentType is not the empty string, it is validated to be either
// image/png or image/jpeg.
func (v *Vino) AddPicture(fh io.Reader, contentType string) error {
	switch contentType {
	case "", "image/jpeg", "image/png":
	default:
		return fmt.Errorf("unexpected content type for image: %q", contentType)
	}

	img, _, err := image.Decode(fh)
	if err != nil {
		return err
	}

	v.Picture = img

	return nil
}

func (v Vino) String() string {
	return fmt.Sprintf("{Name: %q, Rating: %d}", v.Name, v.Rating)
}

func (v *Vino) Store(ctx context.Context, db *bolt.DB) error {
	// If the ID is all-zeros, it's uninitialized and this is a new entry
	needID := true
	for _, v := range v.ID {
		if v != 0 {
			needID = false
			break
		}
	}

	if needID {
		var err error
		v.ID, err = uuid.NewRandom()
		if err != nil {
			return err
		}
	}

	// Encode scaled image as PNG, will contain image data if there is any
	var img bytes.Buffer

	if v.Picture != nil {
		// Scale image down and encode as PNG
		// Get aspect ratio of incoming picture
		bounds := v.Picture.Bounds()
		aspect := float64(bounds.Max.X) / float64(bounds.Max.Y)

		const destHeight = 800

		rect := image.Rect(0, 0, int(destHeight*aspect), destHeight)

		log.Println("bounds:", bounds, "-> aspect:", aspect, "new size:", rect)

		scaled := image.NewRGBA(rect)

		draw.ApproxBiLinear.Scale(scaled, rect, v.Picture, v.Picture.Bounds(), draw.Over, nil)
		err := png.Encode(&img, scaled)
		if err != nil {
			return err
		}
	}

	return db.Update(func(tx *bolt.Tx) error {
		wines, err := tx.CreateBucketIfNotExists([]byte("wines"))
		if err != nil {
			return err
		}

		wine, err := wines.CreateBucketIfNotExists([]byte(v.ID.String()))
		if err != nil {
			return err
		}

		err = wine.Put([]byte("name"), []byte(v.Name))
		if err != nil {
			return err
		}

		err = wine.Put([]byte("rating"), []byte(strconv.Itoa(v.Rating)))
		if err != nil {
			return err
		}

		if img.Len() != 0 {
			err = wine.Put([]byte("picture"), img.Bytes())
			if err != nil {
				return err
			}
		}

		return nil
	})
}

func (v *Vino) StoreComment(ctx context.Context, db *bolt.DB, text string) error {
	return db.Update(func(tx *bolt.Tx) error {
		allWines := tx.Bucket([]byte("wines"))
		if allWines == nil {
			return errNotFound
		}

		wine := allWines.Bucket([]byte(v.ID.String()))
		if wine == nil {
			return errNotFound
		}

		comments, err := wine.CreateBucketIfNotExists([]byte("comments"))
		if err != nil {
			return err
		}

		id, err := uuid.NewRandom()
		if err != nil {
			return err
		}

		return comments.Put(id[:], []byte(text))
	})
}
